#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <signal.h>
#include <termios.h>
#include <unistd.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define MAIN_C
#include "exitfail.h"

#define SERIAL_BAUDRATE B115200
#define BUF_SIZE 2048

static int serial_fd = -1;
static struct termios old_tio;

static int terminated = 0;

#define BASENAME(p) ((strrchr((p), '/') ? : ((p) - 1)) +1)
#define ARRAY_SIZE(a) (sizeof(a) /sizeof((a)[0]))

#define __unused __attribute__((unused))

static void terminate_sig_handler(__unused int signum)
{
	terminated = 1;
}

static void set_sig_handler(int sig_list[], ssize_t num, __sighandler_t handler)
{
	struct sigaction sa;
	int i;
	
	memset(&sa, 0, sizeof(sa));
	sa.sa_handler = handler;

	for(i=0; i<num; i++)
		if(sigaction(sig_list[i], &sa, NULL) < 0)
			exitfail_errno("sigaction");
}

static void restore_serial(void)
{
	int ret;

	ret = tcsetattr(serial_fd, TCSANOW, &old_tio);
	if(ret < 0) exitfail_errno("tcsetattr");
}

static void setup_serial(int fd)
{
	struct termios tio;
	int ret;

	ret = tcgetattr(fd, &old_tio);
	if(ret) exitfail_errno("tcsgetattr");

	if(atexit(restore_serial)) exitfail_errno("atexit");

	memset(&tio, 0, sizeof(tio));
	tio.c_iflag = IGNBRK | IGNPAR;
	tio.c_cflag = CS8 | CLOCAL | CREAD;
	tio.c_cc[VTIME] = 0;
	tio.c_cc[VMIN] = 1;
	ret = cfsetspeed(&tio, SERIAL_BAUDRATE);
	if(ret < 0) exitfail_errno("cfsetspeed");

	ret = tcflush(fd, TCIFLUSH);
	if(ret < 0) exitfail_errno("tcflush");

	ret = tcsetattr(fd, TCSANOW, &tio);
	if(ret) exitfail_errno("tcsetattr");
}

int main(int argc, char *argv[])
{
	int terminate_sig_list[] = {
		SIGHUP, SIGINT, SIGQUIT, SIGPIPE, SIGTERM
	};
	ssize_t ret, len;
/**** sample: variable(mainly oki920): begin ****/
       unsigned char buf[BUF_SIZE];
	unsigned char buf1[BUF_SIZE];
       int cnt_head;
	int cnt_esp;
	int data_len;
	int i;
	ssize_t analen;
	FILE *fp;
	unsigned char proc_id;
/**** sample: variable(mainly oki920): end ****/
/**** sample: variable(sensu1501): begin ****/
//	int fc_flag;
//	int cnt_data;
	unsigned int id;
	unsigned int data_dir;
	unsigned int data_tmp;
	unsigned int data_hm;
	unsigned int data_wind;
	unsigned int data_gust;
	unsigned int data_rain;
	unsigned int data_uvi;
	unsigned int data_light;
	unsigned int data_sum;
/**** sample: variable(sensu1501): end ****/

	exitfail_init();

	if(argc < 2) {
		printf("Usage: %s <device>\n", BASENAME(argv[0]));
		return EXIT_SUCCESS;
	}

	serial_fd = open(argv[1], O_RDWR | O_NOCTTY);
	if(serial_fd < 0) exitfail_errno("open");

	set_sig_handler(terminate_sig_list, ARRAY_SIZE(terminate_sig_list),
			terminate_sig_handler);

	setup_serial(serial_fd);

/**** sample: initial value(oki920): begin ****/
	cnt_head = 0;
	cnt_esp = 0;
	data_len = 0;
	proc_id = 0x00;
/**** sample: initial value(oki920): end ****/
/**** sample: initial value(sensu1501): begin ****/
//	fc_flag = 0;
//	cnt_data = 0;

	id = 0x00;
	data_dir = 0;
	data_tmp = 0;
	data_hm = 0;
	data_wind = 0;
	data_gust = 0;
	data_rain = 0;
	data_uvi = 0;
	data_light = 0;
	data_sum = 0;
/**** sample: initial value(sensu1501): end ****/

	while(!terminated) {
		ret = read(serial_fd, buf, BUF_SIZE);
		if(ret < 0) {
			if(errno == EINTR) continue;
			exitfail_errno("read");
		}
		len = ret;

/**** sample: receive: begin ****/
		for(analen = 0; analen < len; analen++) {
			if(cnt_head == 0 && buf[analen] == 0x7e) {
				buf1[cnt_esp] = buf[analen];
				printf("0x%02x ",  buf1[cnt_esp]);
				cnt_esp++;
				cnt_head++;
			}
			else if(cnt_head == 1) {
				if(buf[analen] == 0x7e) {
					/**** counter reset for sync shift ****/
					cnt_head = 0;
					cnt_esp = 0;
					data_len = 0;

					buf1[cnt_esp] = buf[analen];
					printf("0x%02x ",  buf1[cnt_esp]);
					cnt_esp++;
					cnt_head++;
				}
				else {
					buf1[cnt_esp] = buf[analen];
					proc_id = buf[analen];
					printf("0x%02x ",  buf1[cnt_esp]);
					cnt_esp++;
					cnt_head++;
				}
			}
			else if(cnt_head == 2) {
				buf1[cnt_esp] = buf[analen];
				printf("0x%02x ",  buf1[cnt_esp]);
				cnt_esp++;
				cnt_head++;
			}
			else if(cnt_head == 3) {
				buf1[cnt_esp] = buf[analen];
				printf("0x%02x ",  buf1[cnt_esp]);
				cnt_esp++;
				cnt_head++;
				data_len = buf1[2] * 256 + buf1[3];
			}
			else if(cnt_head > 3 && cnt_esp <= (3 + data_len + 1)) {
				buf1[cnt_esp] = buf[analen];
				printf("0x%02x ",  buf1[cnt_esp]);
				if(cnt_esp == (3 + data_len + 1)) {
/**** sample: get SenSu data: begin ****/
					if(buf1[4] == 0x24 && data_len == 17) {
						for(i = 0; i < 16; i++) data_sum += buf1[4 + i];
						if(buf1[4 + 16] == (data_sum & 0x000000ff)){
/* calculate data*/
							id = buf1[4 + 1];
/**** sample: read data to each sensor data: begin ****/
							data_dir = (buf1[4 + 3] >> 7) * 256 + buf1[4 + 2];
							data_tmp = (buf1[4 + 3] & 0x07) * 256 + buf1[4 + 4];
							data_hm =  buf1[4 + 5];
							data_wind = ((buf1[4 + 3] & 0x10) >> 4) * 256 + buf1[4 + 6];
							data_gust = buf1[4 + 7];
							data_rain = buf1[4 + 8] * 256 + buf1[4 + 9];
							data_uvi = buf1[4 + 10] * 256 + buf1[4 + 11];
							data_light = buf1[4 + 12] * 256 * 256 + buf1[4 + 13] * 256 + buf1[4 + 14];
							data_sum = buf1[4 + 16];
/**** sample: read data to each sensor data: end ****/

/**** sample: sensor data to stdout: begin ****/
							printf("\n");
							printf("0x%02x\n", id);
							printf("0x%03x, %d[degree]\n", data_dir, data_dir);
							printf("0x%03x, %.1f[degree]\n", data_tmp, ((float)data_tmp - 400) / 10);
							printf("0x%02x, %d[%%]\n", data_hm, data_hm);
							printf("0x%02x, %.3f[m/s]\n", data_wind, (float)data_wind * 0.137);
							printf("0x%02x, %.3f[m/s]\n", data_gust, (float)data_gust * 1.114);
							printf("0x%04x, %.1f[mm]\n", data_rain, (float)data_rain * 0.3);
							printf("0x%04x, %d[uw/cm^2]\n", data_uvi, data_uvi);
							printf("0x%06x, %.1f[lux]\n", data_light, (float)data_light / 10);
							printf("0x%02x\n", data_sum);
/**** sample: sensor data to stdout: end ****/

							fp = fopen("/root/log_sensu.txt", "w");

/**** sample: sensor data to log file for post to heroku: begin ****/
							fprintf(fp, "id=0x%02x&", id);
							fprintf(fp, "w_dir=%d&", data_dir);
							fprintf(fp, "tmp=%.1f&", ((float)data_tmp - 400) / 10);
							fprintf(fp, "hum=%d&", data_hm);
							fprintf(fp, "w_speed=%.3f&", (float)data_wind * 0.137);
							fprintf(fp, "w_gust=%.3f&", (float)data_gust * 1.114);
							fprintf(fp, "rain=%.1f&", (float)data_rain * 0.3);
							fprintf(fp, "uvi=%d&", data_uvi);
							fprintf(fp, "lux=%.1f\n", (float)data_light / 10);
/**** sample: sensor data to log file for post to heroku: end ****/

							fclose(fp);
						}
						exit(0);
					}
/**** sample: get SenSu data: end ****/
				}
				else cnt_esp++;
			}
			else {
				cnt_head = 0;
				cnt_esp = 0;
				data_len = 0;
			}
		}
/**** sample: receive: end ****/
	}
	return EXIT_SUCCESS;
}

